﻿# -*- coding=UTF-8 -*-
# Urban Dictionary Wordlist Generator
# V1.2 (2015-07-20 09:02)
# 作者: firefly(dance.of.firefly@gmail.com)
#
# 作用：分析抓取UrbanDictionary网站上的所有词汇列表，保存起来（每个字母一个文件，每行一个词汇），
#       便于后续使用UD_make_mdx.py按词汇批量抓取解释和例句。
# 环境：Python 3.4.3, BeautifulSoup 4.4, Requests 2.7
# 注： 受 https://github.com/natemccoy/udwg 启发，并在此基础上修改优化（支持多进程）。
#
#   Usage
#       At this point, the program is command line only, but there are plans to create a GUI in the future.
#       This program is written in Python and is very easy to use, just simply type:
#       $> python uswg.py
#       And all words from Urban Dictionary will be printed out to STDOUT
#       
#       if you would like just one letter, type:
#       $> python uswg.py A
#       to get all the words that start with the letter 'A':
#       
#       if you would like a range, say 'B' through 'D', type:
#       $> python uswg.py B D
#       
#       Finally, if you would like words that what with special characters or numbers, use the '*' character, type:
#       $> python uswg.py \*
#       
#       Keep in mind that typing:
#       $> python uswg.py A \*
#       is Equivalent to typing:
#       $> python uswg.py
#       
# Refined by firefly to make it work like a charm! (support multiprocessing & error processing!)

#http://www.urbandictionary.com/yesterday.php?date=1999-12-09
import calendar
import sys
import re
import os
import time
import requests
from os import path
from bs4 import BeautifulSoup
from multiprocessing import Pool

################################################################################
# globals
#
################################################################################
#END_DATE = (2007,1,20)
WL_FOLDER = 'WordLists'
UNIQUE_WL_FILE = 'unique_wl.txt'

################################################################################
# dump error information
#
# tries to write error information into file
# Error info contains: when, which letter & page & exception message
################################################################################
def dumpError(date, page, info):
    fname='error.log'
    fpath = ''.join([os.getcwd(), path.sep, fname])
    fw = open(fpath, 'a', encoding='utf-8')  # open or create file
    timestamp = time.asctime( time.localtime(time.time()) )
    data = '---------------\n'+ timestamp +'\nDate: "' + date+'",page: "' +page +'",info: "' + info +'"\n'
    try:
        fw.write(data)
    finally:
        fw.close()


################################################################################
# dump success information
#
# tries to write success information into file
# Success info contains: time, success message
################################################################################
def dumpSuccess(info):
    fname='Wordlists_success.txt'
    fpath = ''.join([os.getcwd(), path.sep, fname])
    fw = open(fpath, 'a', encoding='utf-8')  # open or create file
    timestamp = time.asctime( time.localtime(time.time()) )
    data = timestamp +'\n' + info +'\n'
    try:
        fw.write(data)
    finally:
        fw.close()
        
################################################################################
# 请求页面，分析，获取数据
#
# return page number total-counts when 'page' parameter is None(default)
# otherwise it returns None
################################################################################
def getData(file, date, page=None):
    url = "http://www.urbandictionary.com/yesterday.php?date="+date
    if page:
        url = url + "&page=" +page
        print('---正在处理第 ' + page +' 个页面: ', end = '')
    else:
        print('=====开始处理字符 ' + date + ' 下的词语: ', end = '')
    headers = {
        'Connection': 'Keep-Alive',
        'Accept': 'text/html, application/xhtml+xml, */*',
        'Accept-Language': 'en-US,en;q=0.8,zh-Hans-CN;q=0.5,zh-Hans;q=0.3',
        'User-Agent': 'Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.124 Safari/537.36'
        }
    times =5   #设置网页下载失败后重试的次数
    for i in range(0, times):
        try:
            response = requests.post(url, headers=headers, timeout = 5+i*10)    #每次重试的超时时间依次递增
            soup = BeautifulSoup(response.text, "lxml")
            if soup :
                if i >0 :
                    print('\n第 '+ str(i) +'/' + str(times) +' 次重试成功！')
                break
            else:
                raise IOError("已下载的网页有问题！")
        except Exception as e:
            print(e)
            if i == (times - 1) :
                print('下载页面已达到最大重试次数!!!')
                dumpError(date, str(page) ,str(e))  #写入错误信息
                return None
            else:
                print('准备第 '+ str(i+1) +'/' + str(times) +' 次重试')
                
    # 未指定页面，获取页面总数
    num = 1
    if page == None :  
        pg_indicator = [a.attrs.get('href') for a in soup.select('li a[href^=/yesterday.php?date=]')]
        pattern = re.compile('.*'+date+'&page=(\d+)', re.I) 
        for p_i in pg_indicator :
            pm = pattern.match(p_i)
            if pm==None:
                continue
            else :
                if int(pm.group(1))>num :
                    num = int(pm.group(1))
        print('共 ' + str(num) + ' 个页面====='+str(num))
    
    # 获取页面上的词组
    lst = [a.get_text('href') for a in soup.select('li a[href^=/define.php?term=]')]
    if page == None :
        print('---正在处理第 1 个页面: ', end = '')
    print('共 ' + str(len(lst)) + ' 个词语---'+str(len(lst)))
    if len(lst)>0 :
        try:
            for w in lst:
                w = w.strip() # 默认首尾删除空白符（包括'\n', '\r',  '\t',  ' ')
                if w :
                    file.write(w + '\n')
        except Exception as e:
            print(e)
            
    if num:
        return num
    else:
        return None


################################################################################
# 下载单个日期下的所有词语
################################################################################
def task(date):
    fname = date + ".txt"
    global WL_FOLDER
    fpath = ''.join([os.getcwd(), path.sep, WL_FOLDER, path.sep, fname])
    if os.path.exists(fpath) :
        return
#        os.remove(fpath)  # already exists
    file = open(fpath, 'a', encoding='utf-8')  # open or create file in append mode
    
    pageCount = getData(file, date)
    if pageCount > 1:
        for p in range(2,pageCount+1):
            getData(file, date, str(p))
    file.close()

################################################################################
# 多进程下载特定范围内的多个字母下的词语
#
# 每个进程负责下载一个字母下的所有词语
################################################################################
def udwg_m(start, end):
    dates = getDates(start, end)
#    len(dates)  
    numOfProcs = 30 #下载进程数
    pool = Pool(numOfProcs)
    pool.map(task, dates)

################################################################################
# getDates returns list
#
# returns a list of corresponding dates between 'start' and 'end' in format of "yyyy-mm-dd"
# e.g. "1997-07-01"
################################################################################
def getDates(start, end) :
    year = start[0]
    month = start[1]
    day = start[2]
    
    dates = []
    
    for y in range(year, end[0]+1):
        if y==end[0] :
            months_in_year = end[1]
        else :
            months_in_year = 12
        for m in range(month,months_in_year+1):
            if y==end[0] and m==end[1] :
                days_in_month = end[2]
            else :
                days_in_month = calendar.monthrange(y, m)[1]
            for d in range(day,days_in_month+1):
                dates.append("%d-%02d-%02d" %(y, m, d))
            day = 1
        month =1
        y +=1
    return dates


################################################################################
# 读取单个文件内的单词，每个单词去除无效空白、转小写后分别加入集合
#
# 返回值：各元素互不相同的集合
################################################################################
def readLinesAsCollection(wl_file_path):
    wl_file = open(wl_file_path, 'r', encoding='utf-8')  # open or create file
    collection = set()
    for line in wl_file:
        line = line.strip() # 默认首尾删除空白符（包括'\n', '\r',  '\t',  ' ')
        line = line.lower() # 所有字母全部小写
        collection.add(line)
    wl_file.close()
    return collection
    
################################################################################
# 多进程读取文件夹下所有文件的单词，过滤去重复后写入 UNIQUE_WL_FILE 内。
#
################################################################################
def filterWordsInFolder():
    folder_path = ''.join([os.getcwd(), path.sep, WL_FOLDER])
    files = [(folder_path + path.sep + f) for f in os.listdir(folder_path) if os.path.isfile(folder_path + path.sep + f) ]
    numOfProcs = 20  #处理分配的进程数
    if (numOfProcs!=0) :
        pool = Pool(numOfProcs)
        cols_list = pool.map(readLinesAsCollection, files)
    else :
        cols_list = []
    unique_file_path = ''.join([folder_path, path.sep, UNIQUE_WL_FILE])
    unique_file = open(unique_file_path,  'w', encoding='utf-8')  # open or create file in append mode
    cols = set()
    for c_l in cols_list :
        cols |= c_l
    for c in cols :
        unique_file.write(c+'\n')
    unique_file.close()

################################################################################
# __main__
#
# checks arguments and calls udwg() with corresponding arguments
# if the arguments are faulty, the program will exit with a message
################################################################################
if __name__ == "__main__":
    argc = len(sys.argv)
    start = time.time() 
    
    START_DATE = (1999, 12, 9)
    tm = time.localtime(time.time())
    END_DATE = (tm[0], tm[1], tm[2])
#    END_DATE = (1999,12,11)
    
    udwg_m(START_DATE, END_DATE)
    filterWordsInFolder()
#    task("2015-09-09")
    
    end = time.time() 
    print("耗时: %s 秒！" % (end-start))
    dumpSuccess("耗时: "+str(end-start) +" 秒！")  